#include <amxmodx>
#include <fakemeta>
#include <fakemeta_util>
#include <hamsandwich>
#include <xs>
#include <zombieplague_x>

#define PLUGIN "[ZP] Extra: CSO Weapon"
#define VERSION "1.0"
#define AUTHOR "KORD_12.7 | qubka"

#pragma ctrlchar '\'

//**********************************************
//* Weapon Settings.                           *
//**********************************************

// Main
#define WEAPON_NAME 				"weapon_chainsaw_zpx"
#define WEAPON_REFERANCE			"weapon_m249"
#define WEAPON_MAX_CLIP				100
#define WEAPON_DEFAULT_AMMO			250

#define WEAPON_TIME_NEXT_IDLE 		5.0
#define WEAPON_TIME_NEXT_ATTACK 	0.075
#define WEAPON_TIME_DELAY_DEPLOY 	1.23
#define WEAPON_TIME_DELAY_RELOAD 	3.0

// Damage
#define CHAINSAW_DAMAGE				100.0
#define CHAINSAW_DAMAGE2			500.0

// Attack distance
#define CHAINSAW_DISTANCE			65.0
#define CHAINSAW_DISTANCE2			70.0

// Attack knock
#define CHAINSAW_KNOCK				200.0
#define CHAINSAW_KNOCK2				500.0

// Models
#define MODEL_WORLD					"models/zombie_plague/w_chainsaw.mdl"
#define MODEL_VIEW					"models/zombie_plague/v_chainsaw.mdl"
#define MODEL_PLAYER				"models/zombie_plague/p_chainsaw.mdl"

// Sounds
new const SAW_SOUND[3][] =
{
	"weapons/chainsaw_attack1_loop.wav",
	"weapons/chainsaw_hit1.wav",
	"weapons/chainsaw_hit4.wav"
}

// Sprites
#define WEAPON_HUD_TXT				"sprites/weapon_chainsaw_zpx.txt"
#define WEAPON_HUD_SPR_1			"sprites/zpx/640hud21.spr"
#define WEAPON_HUD_SPR_2			"sprites/zpx/640hud84.spr"

// Animation
#define ANIM_EXTENSION				"chainsaw"

// Animation sequences
enum
{	
	ANIM_IDLE,
	ANIM_DRAW,
	ANIM_DRAW_EMPTY,
	ANIM_ATTACK_BEGIN,
	ANIM_ATTACK_LOOP,
	ANIM_ATTACK_END,
	ANIM_RELOAD,
	ANIM_SLASH1,
	ANIM_SLASH2,
	ANIM_SLASH3,
	ANIM_SLASH4,
	ANIM_IDLE_EMPTY
};

//**********************************************
//* Some macroses.                             *
//**********************************************

#define MDLL_Spawn(%0)			dllfunc(DLLFunc_Spawn, %0)
#define MDLL_Touch(%0,%1)		dllfunc(DLLFunc_Touch, %0, %1)

#define SET_MODEL(%0,%1)		engfunc(EngFunc_SetModel, %0, %1)
#define SET_ORIGIN(%0,%1)		engfunc(EngFunc_SetOrigin, %0, %1)

#define PRECACHE_MODEL(%0)		engfunc(EngFunc_PrecacheModel, %0)
#define PRECACHE_SOUND(%0)		engfunc(EngFunc_PrecacheSound, %0)
#define PRECACHE_GENERIC(%0)		engfunc(EngFunc_PrecacheGeneric, %0)

#define MODEL_INDEX(%0) engfunc(EngFunc_ModelIndex,%0)

#define MESSAGE_BEGIN(%0,%1,%2,%3)	engfunc(EngFunc_MessageBegin, %0, %1, %2, %3)
#define MESSAGE_END()			message_end()

#define WRITE_ANGLE(%0)			engfunc(EngFunc_WriteAngle, %0)
#define WRITE_BYTE(%0)			write_byte(%0)
#define WRITE_COORD(%0)			engfunc(EngFunc_WriteCoord, %0)
#define WRITE_STRING(%0)		write_string(%0)
#define WRITE_SHORT(%0)			write_short(%0)

#define INSTANCE(%0)			((%0 == -1) ? 0 : %0)

#define BitSet(%0,%1) (%0 |= (1 << (%1 - 1)))
#define BitClear(%0,%1) (%0 &= ~(1 << (%1 - 1)))
#define BitCheck(%0,%1) (%0 & (1 << (%1 - 1)))

//**********************************************
//* PvData Offsets.                            *
//**********************************************

// Linux extra offsets
#define extra_offset_weapon		4
#define extra_offset_player		5

new g_bitIsConnected;
#define m_rgpPlayerItems_CWeaponBox	34

#define HIDEHUD_CROSSHAIR (1 << 6)

// CBasePlayerItem
#define m_pPlayer			41
#define m_pNext				42

// CBasePlayerWeapon
#define m_flNextPrimaryAttack		46
#define m_flNextSecondaryAttack		47
#define m_flTimeWeaponIdle		48
#define m_iPrimaryAmmoType		49
#define m_iClip					51
#define m_fInReload				54
#define m_iDirection			60
#define m_flAccuracy			62
#define m_iShotsFired			64
#define m_fWeaponState			74
#define m_LastHitGroup 			75
#define m_iChargeReady			55
#define m_flLastEventCheck  	38

// CBaseMonster
#define m_flNextAttack			83

// CBasePlayer
#define m_iFOV							363
#define m_iHideHUD						361
#define m_iLastZoom 					109
#define m_fResumeZoom      				110
#define m_rgpPlayerItems_CBasePlayer	367
#define m_pActiveItem					373
#define m_rgAmmo_CBasePlayer			376
#define m_szAnimExtention				492

#define IsValidPev(%0) (pev_valid(%0) == 2)

#define IsObserver(%0) pev(%0,pev_iuser1)

#define OBS_IN_EYE 4

new g_iszWeaponKey;

#define IsCustomItem(%0) (pev(%0, pev_impulse) == g_iszWeaponKey)


#define THINK_ATTACK		5324233

#define set_think(%0,%1,%2)			\
							\
	remove_task(%0+THINK_ATTACK); 	\
	set_task(%2,%1,%0+THINK_ATTACK)
	
//**********************************************
//* Let's code our weapon.                     *
//**********************************************

new SprPuff;
new iBlood[2];

Weapon_OnPrecache()
{
	PRECACHE_MODEL(MODEL_VIEW);
	PRECACHE_MODEL(MODEL_PLAYER);
	PRECACHE_MODEL(MODEL_WORLD) ;
	
	new i;
	for(i = 0; i < sizeof(SAW_SOUND); i++)
		PRECACHE_SOUND(SAW_SOUND[i])
		
	PRECACHE_SOUND("weapons/chainsaw_idle.wav");
	PRECACHE_SOUND("weapons/chainsaw_draw.wav");
	PRECACHE_SOUND("weapons/chainsaw_draw1.wav");
	PRECACHE_SOUND("weapons/chainsaw_attack1_start.wav");
	PRECACHE_SOUND("weapons/chainsaw_attack1_end.wav");
	PRECACHE_SOUND("weapons/chainsaw_reload.wav");
	PRECACHE_SOUND("weapons/chainsaw_slash1.wav");
	PRECACHE_SOUND("weapons/chainsaw_slash2.wav");
	PRECACHE_SOUND("weapons/chainsaw_slash3.wav");
	PRECACHE_SOUND("weapons/chainsaw_slash4.wav");
	
	SprPuff = PRECACHE_MODEL("sprites/wall_puff1.spr");
	
	iBlood[0] = PRECACHE_MODEL("sprites/bloodspray.spr");
	iBlood[1] = PRECACHE_MODEL("sprites/blood.spr");
	
	PRECACHE_GENERIC(WEAPON_HUD_TXT);
	PRECACHE_GENERIC(WEAPON_HUD_SPR_1);
	PRECACHE_GENERIC(WEAPON_HUD_SPR_2);
}

Weapon_OnSpawn(const iItem)
{
	// Setting world model.
	SET_MODEL(iItem, MODEL_WORLD);
}

Weapon_OnDeploy(const iItem, const iPlayer, const iClip, const iAmmoPrimary, const iChargeReady)
{
	#pragma unused iClip, iAmmoPrimary, iChargeReady
	
	static iszViewModel;
	if (iszViewModel || (iszViewModel = engfunc(EngFunc_AllocString, MODEL_VIEW)))
	{
		set_pev_string(iPlayer, pev_viewmodel2, iszViewModel);
	}
	
	static iszPlayerModel;
	if (iszPlayerModel || (iszPlayerModel = engfunc(EngFunc_AllocString, MODEL_PLAYER)))
	{
		set_pev_string(iPlayer, pev_weaponmodel2, iszPlayerModel);
	}
	
	// Cancel any reload in progress.
	set_pdata_int(iItem, m_fInReload, 0, extra_offset_weapon);

	remove_task(iPlayer+THINK_ATTACK);
	
	//Hide CrossHair
	message_begin(MSG_ONE, get_user_msgid("HideWeapon"), _, iPlayer);
	write_byte(HIDEHUD_CROSSHAIR);
	message_end();
	
	set_pdata_float(iItem, m_flTimeWeaponIdle, WEAPON_TIME_DELAY_DEPLOY, extra_offset_weapon);
	set_pdata_float(iPlayer, m_flNextAttack, WEAPON_TIME_DELAY_DEPLOY, extra_offset_player);
	set_pdata_int(iItem, m_iChargeReady, 0, extra_offset_weapon);

	// Deploy with default weapon settings
	Weapon_DefaultDeploy(iItem, iPlayer, MODEL_VIEW, MODEL_PLAYER, ANIM_DRAW, ANIM_EXTENSION);
}

Weapon_OnHolster(const iItem, const iPlayer, const iClip, const iAmmoPrimary, const iChargeReady)
{
	#pragma unused iPlayer, iClip, iAmmoPrimary, iChargeReady
	
	Weapon_EndAttack(iItem, iPlayer);
	
	remove_task(iPlayer+THINK_ATTACK);
	
	//Show CrossHair
	message_begin(MSG_ONE, get_user_msgid("HideWeapon"), _, iPlayer);
	write_byte(1>>6);
	message_end();
	
	// Cancel any reload in progress.
	set_pdata_int(iItem, m_fInReload, 0, extra_offset_weapon);
}

Weapon_OnIdle(const iItem, const iPlayer, const iClip, const iAmmoPrimary, const iChargeReady)
{
	#pragma unused iClip, iAmmoPrimary, iChargeReady
	
	ExecuteHamB(Ham_Weapon_ResetEmptySound, iItem);

	if(get_pdata_int(iItem, m_iChargeReady, extra_offset_weapon))
	{
		Weapon_EndAttack(iItem, iPlayer);
		return;
	}
	
	if (get_pdata_int(iItem, m_flTimeWeaponIdle, extra_offset_weapon) > 0.0)
	{
		return;
	}

	set_pdata_float(iItem, m_flTimeWeaponIdle, WEAPON_TIME_NEXT_IDLE, extra_offset_weapon);
	Weapon_SendAnim(iPlayer,  !iClip ? ANIM_IDLE_EMPTY : ANIM_IDLE);
}

Weapon_OnReload(const iItem, const iPlayer, const iClip, const iAmmoPrimary, const iChargeReady)
{
	#pragma unused iAmmoPrimary, iChargeReady
	
	if (min(WEAPON_MAX_CLIP - iClip, iAmmoPrimary) <= 0)
	{
		return;
	}
	
	if(iChargeReady > 0)
	{
		Weapon_EndAttack(iItem, iPlayer);
		return;
	}
	
	if (get_pdata_int(iItem, m_flNextSecondaryAttack, extra_offset_weapon) > 0.0)
	{
		return;
	}
	
	set_pdata_int(iItem, m_iClip, 0, extra_offset_weapon);
	
	ExecuteHam(Ham_Weapon_Reload, iItem);
	
	static szAnimation[64];
	formatex(szAnimation, charsmax(szAnimation), "ref_reload_%s", ANIM_EXTENSION);
	Player_SetAnimation(iPlayer, szAnimation);
	
	set_pdata_int(iItem, m_iClip, iClip, extra_offset_weapon);
	
	set_pdata_float(iPlayer, m_flNextAttack, WEAPON_TIME_DELAY_RELOAD, extra_offset_player);
	set_pdata_float(iItem, m_flTimeWeaponIdle, WEAPON_TIME_DELAY_RELOAD, extra_offset_weapon);
	set_pdata_float(iItem, m_flNextSecondaryAttack, WEAPON_TIME_DELAY_RELOAD, extra_offset_weapon);
	
	Weapon_SendAnim(iPlayer, ANIM_RELOAD);
}

Weapon_OnPrimaryAttack(const iItem, const iPlayer, const iClip, const iAmmoPrimary, const iChargeReady)
{
	#pragma unused iAmmoPrimary, iChargeReady
	
	if(pev(iPlayer, pev_waterlevel) == 3)
	{
		if(iChargeReady > 0)
		{
			Weapon_EndAttack(iItem, iPlayer);
		}
		return;
	}
	
	if(iClip <= 0)
	{
		if(iChargeReady > 0)
		{
			Weapon_EndAttack(iItem, iPlayer);
		}
		else
		{
			Weapon_OnSecondaryAttack(iItem, iPlayer, iClip, iAmmoPrimary, iChargeReady)
		}
		return;
	}
	
	switch(iChargeReady)
	{
		case 0:
		{
			Weapon_SendAnim(iPlayer, ANIM_ATTACK_BEGIN);
			
			set_pdata_int(iItem, m_iChargeReady, 1, extra_offset_weapon);
			
			set_pdata_float(iItem, m_flTimeWeaponIdle, 0.4, extra_offset_weapon);
			set_pdata_float(iItem, m_flNextPrimaryAttack, 0.4, extra_offset_weapon);
			set_pdata_float(iItem, m_flNextSecondaryAttack, 0.4, extra_offset_weapon);
			return;
		}
		case 1:
		{
			static Float: vecVelocity[3];
		
			if(get_pdata_int(iItem, m_flTimeWeaponIdle, extra_offset_weapon) > 0.0)
			{
				return;
			}
			
			pev(iPlayer, pev_velocity, vecVelocity);
			
			set_pdata_int(iItem, m_iClip, iClip - 1, extra_offset_weapon);
			
			Weapon_SendAnim(iPlayer, ANIM_ATTACK_LOOP);

			emit_sound(iPlayer, CHAN_WEAPON, SAW_SOUND[0], 0.9, ATTN_NORM, 0, PITCH_NORM);
			
			static iFlags, szAnimation[64];
			iFlags = pev(iPlayer, pev_flags);
			
			formatex(szAnimation, charsmax(szAnimation), "ref_shoot_%s", ANIM_EXTENSION);
			Player_SetAnimation(iPlayer, szAnimation);

			set_pdata_float(iItem, m_flTimeWeaponIdle, WEAPON_TIME_NEXT_ATTACK, extra_offset_weapon);
			set_pdata_float(iItem, m_flNextPrimaryAttack, WEAPON_TIME_NEXT_ATTACK, extra_offset_weapon);
			set_pdata_float(iItem, m_flNextSecondaryAttack, WEAPON_TIME_NEXT_ATTACK, extra_offset_weapon);
			
			Weapon_Slash(iItem, iPlayer, 0.0, true);
			
			if (xs_vec_len(vecVelocity) > 0)
			{
				Weapon_KickBack(iItem, iPlayer, 1.5, 0.45, 0.225, 0.05, 6.5, 2.5, 7);
			}
			else if (!(iFlags & FL_ONGROUND))
			{
				Weapon_KickBack(iItem, iPlayer, 2.0, 1.0, 0.5, 0.35, 9.0, 6.0, 5);
			}
			else if (iFlags & FL_DUCKING)
			{
				Weapon_KickBack(iItem, iPlayer, 0.9, 0.35, 0.15, 0.025, 5.5, 1.5, 9);
			}
			else
			{
				Weapon_KickBack(iItem, iPlayer, 1.0, 0.375, 0.175, 0.0375, 5.75, 1.75, 8);
			}
		}
	}
}

// Secondary attack: slow, low damage, bigger attack distance, no ammo required
public Weapon_OnSecondaryAttack(const iItem, const iPlayer, const iClip, const iAmmoPrimary, const iChargeReady)
{
	#pragma unused iAmmoPrimary, iChargeReady
	
	if(pev(iPlayer, pev_waterlevel) == 3)
	{
		if(iChargeReady > 0)
		{
			Weapon_EndAttack(iItem, iPlayer);
		}
		return;
	}
	
	if(iChargeReady > 0)
	{
		Weapon_EndAttack(iItem, iPlayer);
		return;
	}
	
	if(get_pdata_int(iItem, m_flNextSecondaryAttack, extra_offset_weapon) > 0.0)
	{
		return;
	}
	
	static szAnimation[64];
	Weapon_SendAnim(iPlayer, !iClip ? ANIM_SLASH4 : ANIM_SLASH2);

	formatex(szAnimation, charsmax(szAnimation), "ref_shoot2_%s", ANIM_EXTENSION);
	Player_SetAnimation(iPlayer, szAnimation);
	
	set_pdata_float(iItem, m_flTimeWeaponIdle, 1.2, extra_offset_weapon);
	set_pdata_float(iItem, m_flNextPrimaryAttack, 1.2, extra_offset_weapon);
	set_pdata_float(iItem, m_flNextSecondaryAttack, 1.2, extra_offset_weapon);
	
	set_think(iPlayer, "Chainsaw_Stab", 0.105);
}

public Chainsaw_Stab(iPlayer)
{
	iPlayer -= THINK_ATTACK;
	
	new iItem;
	
	if (!CheckItem2(iPlayer, iItem))
	{
		return;
	}

	for (new Float: flRightBase = -14.0, i = 0; i < 15; i++)
	{
		Weapon_Slash(iItem, iPlayer, flRightBase += 4.0, false);
	}
}

// Stop attacking
Weapon_EndAttack(const iItem, const iPlayer)
{
	Weapon_SendAnim(iPlayer, ANIM_ATTACK_END);
	
	set_pdata_float(iItem, m_flTimeWeaponIdle, 2.0, extra_offset_weapon);
	set_pdata_float(iItem, m_flNextPrimaryAttack, 2.0, extra_offset_weapon);
	set_pdata_float(iItem, m_flNextSecondaryAttack, 2.0, extra_offset_weapon);
	
	set_pdata_int(iItem, m_iChargeReady, 0, extra_offset_weapon);
}

//**********************************************
//* Main Attack Weapon.    		               *
//**********************************************

public Weapon_Slash(const iItem, const iPlayer, Float: flRightScale, bool: bSlash)
{	
	new Float: Origin[3]; 
	new Float: VecEnd[3];
	new Float: VecScr[3]; 
	
	new Float: flFraction; 
	
	new iTrace;
	new iVictim;
	
	Weapon_GetGunPosition(iPlayer, Origin, VecScr, 0.0, flRightScale, 0.0);

	angle_vector(VecScr, ANGLEVECTOR_FORWARD, VecScr);
	xs_vec_mul_scalar(VecScr, (bSlash ? CHAINSAW_DISTANCE : CHAINSAW_DISTANCE2), VecEnd);
	xs_vec_mul_scalar(VecScr, (bSlash ? CHAINSAW_KNOCK : CHAINSAW_KNOCK2), VecScr);
	xs_vec_add(Origin, VecEnd, VecEnd);
	
	engfunc(EngFunc_TraceLine, Origin, VecEnd, DONT_IGNORE_MONSTERS, iPlayer, (iTrace = create_tr2()));
	get_tr2(iTrace, TR_flFraction, flFraction);
		
	if (flFraction >= 1.0)
	{
		engfunc(EngFunc_TraceHull, Origin, VecEnd, DONT_IGNORE_MONSTERS, HULL_HEAD, iPlayer, iTrace);
		get_tr2(iTrace, TR_flFraction, flFraction);
		
		if (flFraction < 1.0)
		{
			iVictim = INSTANCE(get_tr2(iTrace, TR_pHit));
			
			if (!iVictim || ExecuteHamB(Ham_IsBSPModel, iVictim))
			{
				FindHullIntersection(Origin, iTrace, Float: {-16.0, -16.0, -18.0}, Float: {16.0,  16.0,  18.0}, iPlayer);
			}
		}
	}
	
	get_tr2(iTrace, TR_flFraction, flFraction);
	
	if (flFraction < 1.0)
	{
		iVictim = INSTANCE(get_tr2(iTrace, TR_pHit));
		
		if(iVictim > 0 && pev(iVictim, pev_takedamage) != DAMAGE_NO && pev(iVictim, pev_solid) != SOLID_NOT)
		{	
			if(ExecuteHamB(Ham_IsPlayer, iVictim))
			{
				if(zp_get_user_zombie(iVictim) || zp_get_user_nemesis(iVictim))
				{
				
					VecScr[2] += bSlash ? CHAINSAW_KNOCK : CHAINSAW_KNOCK2;	
					set_pev(iVictim, pev_velocity, VecScr);	

					static iBloodColor;
					
					if ((iBloodColor = ExecuteHamB(Ham_BloodColor, iVictim)) != -1)
					{
						new Float: vecEndPos[3];	
						get_tr2(iTrace, TR_vecEndPos, vecEndPos);	
					
						UTIL_BloodDrips(vecEndPos, iBloodColor, 200.0);
					}
				}
				emit_sound(iVictim, CHAN_VOICE, SAW_SOUND[2], 1.0, ATTN_NORM, 0, PITCH_NORM);
			}
			
			ExecuteHamB(Ham_TakeDamage, iVictim, iItem, iPlayer, (bSlash ? CHAINSAW_DAMAGE : CHAINSAW_DAMAGE2), DMG_BULLET|DMG_NEVERGIB);	
		}
		else
		{
			if(bSlash)
			{
				static TE_FLAG;
				
				get_tr2(iTrace, TR_vecEndPos, VecEnd);
					
				engfunc(EngFunc_MessageBegin, MSG_PAS, SVC_TEMPENTITY, VecEnd, 0);
				write_byte(TE_SPARKS);
				engfunc(EngFunc_WriteCoord, VecEnd[0]);
				engfunc(EngFunc_WriteCoord, VecEnd[1]);
				engfunc(EngFunc_WriteCoord, VecEnd[2]);
				message_end();

				TE_FLAG |= TE_EXPLFLAG_NODLIGHTS
				TE_FLAG |= TE_EXPLFLAG_NOSOUND
				TE_FLAG |= TE_EXPLFLAG_NOPARTICLES

				engfunc(EngFunc_MessageBegin, MSG_PAS, SVC_TEMPENTITY, VecEnd, 0);
				write_byte(TE_EXPLOSION);
				engfunc(EngFunc_WriteCoord, VecEnd[0]);
				engfunc(EngFunc_WriteCoord, VecEnd[1]);
				engfunc(EngFunc_WriteCoord, VecEnd[2] - 5.0);
				write_short(SprPuff);
				write_byte(5);
				write_byte(50);
				write_byte(TE_FLAG);
				message_end();

				engfunc(EngFunc_MessageBegin, MSG_PAS, SVC_TEMPENTITY, VecEnd, 0);
				write_byte(TE_GUNSHOTDECAL);
				engfunc(EngFunc_WriteCoord, VecEnd[0]);
				engfunc(EngFunc_WriteCoord, VecEnd[1]);
				engfunc(EngFunc_WriteCoord, VecEnd[2]);
				write_short(iPlayer);
				write_byte(random_num(41, 45));
				message_end();
			}
			
			emit_sound(iPlayer, CHAN_VOICE, SAW_SOUND[1], 1.0, ATTN_NORM, 0, PITCH_NORM);
		}
	}
	
	free_tr2(iTrace);
}

//*********************************************************************
//*           Don't modify the code below this line unless            *
//*          	 you know _exactly_ what you are doing!!!             *
//*********************************************************************

#define MSGID_WEAPONLIST 78

new g_iItemID;
new g_surv_chain
public plugin_precache()
{
	Weapon_OnPrecache();
	
	g_iszWeaponKey = engfunc(EngFunc_AllocString, WEAPON_NAME);
	register_message(get_user_msgid("DeathMsg"), "MsgHook_Death");
	
	register_message(MSGID_WEAPONLIST, "MsgHook_WeaponList");
	
	g_surv_chain = zp_register_survivor_class("Survivor Chainsaw" , "surv", "survivor", 3000) ;
}

public plugin_init()
{
	register_plugin(PLUGIN, VERSION, AUTHOR);
	
	register_forward(FM_SetModel,		"FakeMeta_SetModel",			false);
	
	g_iItemID = zp_register_extra_item("chainsaw", 0, 0);
	
	// Weaponbox
	RegisterHam(Ham_Spawn, 		"weaponbox", 		"HamHook_Weaponbox_Spawn_Post", true);
	
	// Item (weapon) hooks
	RegisterHam(Ham_Item_Deploy,		WEAPON_REFERANCE, "HamHook_Item_Deploy_Post",	true);
	RegisterHam(Ham_Item_Holster,		WEAPON_REFERANCE, "HamHook_Item_Holster",	false);
	RegisterHam(Ham_Item_AddToPlayer,	WEAPON_REFERANCE, "HamHook_Item_AddToPlayer",	false);
	RegisterHam(Ham_Item_PostFrame,		WEAPON_REFERANCE, "HamHook_Item_PostFrame",	false);
	
	RegisterHam(Ham_Weapon_Reload,		WEAPON_REFERANCE, "HamHook_Item_Reload",		false);
	RegisterHam(Ham_Weapon_WeaponIdle,	WEAPON_REFERANCE, "HamHook_Item_WeaponIdle",	false);
	RegisterHam(Ham_Weapon_PrimaryAttack, WEAPON_REFERANCE, "HamHook_Item_PrimaryAttack",	false);
	
	// Block client weapon here
	register_forward(FM_UpdateClientData,	"FakeMeta_UpdateClientData_Post", true);
}

public zp_extra_item_selected(id, itemid)
{
	if (itemid == g_iItemID)
	{
		Weapon_Give(id);
	}
}

//**********************************************
//* Block client weapon.                       *
//**********************************************

public FakeMeta_UpdateClientData_Post(const iPlayer, const iSendWeapons, const CD_Handle)
{
	static iActiveItem;iActiveItem = get_pdata_cbase(iPlayer, m_pActiveItem, extra_offset_player);
	
	if (!IsValidPev(iActiveItem) || !IsCustomItem(iActiveItem))
	{
		return FMRES_IGNORED;
	}

	set_cd(CD_Handle, CD_flNextAttack, get_gametime() + 0.001);
	return FMRES_IGNORED;
}
//**********************************************
//* Item (weapon) hooks.                       *
//**********************************************

	#define _call.%0(%1,%2) \
								\
	Weapon_On%0						\
	(							\
		%1, 						\
		%2,						\
								\
		get_pdata_int(%1, m_iClip, extra_offset_weapon),	\
								\
		GetAmmoInventory(%2, PrimaryAmmoIndex(%1)),	\
								\
		get_pdata_int(%1, m_iChargeReady, extra_offset_weapon)	\
	) 


public HamHook_Item_Deploy_Post(const iItem)
{
	new iPlayer; 
	
	if (!CheckItem(iItem, iPlayer))
	{
		return HAM_IGNORED;
	}
	
	_call.Deploy(iItem, iPlayer);
	return HAM_IGNORED;
}

public HamHook_Item_Holster(const iItem)
{
	new iPlayer; 
	
	if (!CheckItem(iItem, iPlayer))
	{
		return HAM_IGNORED;
	}
	
	set_pev(iPlayer, pev_viewmodel, 0);
	set_pev(iPlayer, pev_weaponmodel, 0);
	
	_call.Holster(iItem, iPlayer);
	return HAM_SUPERCEDE;
}

public HamHook_Item_WeaponIdle(const iItem)
{
	static iPlayer; 
	
	if (!CheckItem(iItem, iPlayer))
	{
		return HAM_IGNORED;
	}

	_call.Idle(iItem, iPlayer);
	return HAM_SUPERCEDE;
}

public HamHook_Item_Reload(const iItem)
{
	static iPlayer; 
	
	if (!CheckItem(iItem, iPlayer))
	{
		return HAM_IGNORED;
	}
	
	_call.Reload(iItem, iPlayer);
	return HAM_SUPERCEDE;
}

public HamHook_Item_PrimaryAttack(const iItem)
{
	static iPlayer; 
	
	if (!CheckItem(iItem, iPlayer))
	{
		return HAM_IGNORED;
	}
	
	_call.PrimaryAttack(iItem, iPlayer);
	return HAM_SUPERCEDE;
}

public HamHook_Item_PostFrame(const iItem)
{
	static iButton, iPlayer; 
	
	if (!CheckItem(iItem, iPlayer))
	{
		return HAM_IGNORED;
	}
	
	// Complete reload
	if (get_pdata_int(iItem, m_fInReload, extra_offset_weapon))
	{
		new iClip		= get_pdata_int(iItem, m_iClip, extra_offset_weapon); 
		new iPrimaryAmmoIndex	= PrimaryAmmoIndex(iItem);
		new iAmmoPrimary		= GetAmmoInventory(iPlayer, iPrimaryAmmoIndex);
		new iAmount		= min(WEAPON_MAX_CLIP - iClip, iAmmoPrimary);
		
		set_pdata_int(iItem, m_iClip, iClip + iAmount, extra_offset_weapon);
		set_pdata_int(iItem, m_fInReload, false, extra_offset_weapon);

		SetAmmoInventory(iPlayer, iPrimaryAmmoIndex, iAmmoPrimary - iAmount);
	}
	
	// Call secondary attack
	if ((iButton = pev(iPlayer, pev_button)) & IN_ATTACK2 
		&& get_pdata_float(iItem, m_flNextSecondaryAttack, extra_offset_weapon) <= 0.0)
	{
		_call.SecondaryAttack(iItem, iPlayer);
		set_pev(iPlayer, pev_button, iButton & ~IN_ATTACK2);
	}
	
	return HAM_IGNORED;
}

//**********************************************
//* Death Message.                             *
//**********************************************

public MsgHook_Death(Msg_ID, Msg_Dth, iPlayer)
{
	static iItem[33], iAttacker, szTruncatedWeaponName[32], iActiveItem;
	
	get_msg_arg_string(4, iItem, charsmax(iItem));
	
	copy(szTruncatedWeaponName, charsmax(szTruncatedWeaponName), WEAPON_NAME);
	replace(szTruncatedWeaponName, charsmax(szTruncatedWeaponName), "weapon_", "");
	
	iAttacker = get_msg_arg_int(1);

	if(!is_user_alive(iAttacker) || !is_user_connected(iAttacker) || iAttacker == get_msg_arg_int(2))
		return;

	if(equal(iItem, szTruncatedWeaponName))
	{	
		iActiveItem = get_pdata_cbase(iAttacker, m_pActiveItem, extra_offset_player);
	
		if(IsCustomItem(iActiveItem))
			set_msg_arg_string(4, "knife");
	}
}

//**********************************************
//* Kick back.                                 *
//**********************************************

Weapon_KickBack(const iItem, const iPlayer, Float: upBase, Float: lateralBase, const Float: upMod, const Float: lateralMod, Float: upMax, Float: lateralMax, const directionChange)
{
	static iDirection; 
	static iShotsFired; 
	
	static Float: vecPunchangle[3];
	pev(iPlayer, pev_punchangle, vecPunchangle);
	
	if ((iShotsFired = get_pdata_int(iItem, m_iShotsFired, extra_offset_weapon)) != 1)
	{
		upBase += iShotsFired * upMod;
		lateralBase += iShotsFired * lateralMod;
	}
	
	upMax *= -1.0;
	vecPunchangle[0] -= upBase;
 
	if (upMax >= vecPunchangle[0])
	{
		vecPunchangle[0] = upMax;
	}
	
	if ((iDirection = get_pdata_int(iItem, m_iDirection, extra_offset_weapon)))
	{
		vecPunchangle[1] += lateralBase;
		
		if (lateralMax < vecPunchangle[1])
		{
			vecPunchangle[1] = lateralMax;
		}
	}
	else
	{
		lateralMax *= -1.0;
		vecPunchangle[1] -= lateralBase;
		
		if (lateralMax > vecPunchangle[1])
		{
			vecPunchangle[1] = lateralMax;
		}
	}
	
	if (!random_num(0, directionChange))
	{
		set_pdata_int(iItem, m_iDirection, !iDirection, extra_offset_weapon);
	}
	
	set_pev(iPlayer, pev_punchangle, vecPunchangle);
}

//**********************************************
//* Weapon Blodd Drips.    		               *
//**********************************************

UTIL_BloodDrips(Float: vecSpot[3], const iBloodColor, Float: flDamage)
{
	if (flDamage <= 0.0)
	{
		return;
	}

	if (flDamage > 200.0)
	{
		flDamage = 200.0;
	}

	MESSAGE_BEGIN(MSG_PVS, SVC_TEMPENTITY, vecSpot, 0);
	WRITE_BYTE( TE_BLOODSPRITE );
	WRITE_COORD(vecSpot[0]);
	WRITE_COORD(vecSpot[1]);
	WRITE_COORD(vecSpot[2]);
	WRITE_SHORT(iBlood[0]);
	WRITE_SHORT(iBlood[1]);
	WRITE_BYTE(iBloodColor);
	WRITE_BYTE(min(max(3, floatround(flDamage) / 10), 16));
	MESSAGE_END();
}

//**********************************************
//* Create and check our custom weapon.        *
//**********************************************

Weapon_Create(const Float: vecOrigin[3] = {0.0, 0.0, 0.0}, const Float: vecAngles[3] = {0.0, 0.0, 0.0})
{
	new iWeapon;

	static iszAllocStringCached;
	if (iszAllocStringCached || (iszAllocStringCached = engfunc(EngFunc_AllocString, WEAPON_REFERANCE)))
	{
		iWeapon = engfunc(EngFunc_CreateNamedEntity, iszAllocStringCached);
	}
	
	if (!IsValidPev(iWeapon))
	{
		return FM_NULLENT;
	}
	
	MDLL_Spawn(iWeapon);
	SET_ORIGIN(iWeapon, vecOrigin);
	
	set_pdata_int(iWeapon, m_iClip, WEAPON_MAX_CLIP, extra_offset_weapon);

	set_pev_string(iWeapon, pev_classname, g_iszWeaponKey );
	set_pev(iWeapon, pev_impulse, g_iszWeaponKey);
	set_pev(iWeapon, pev_angles, vecAngles);
	
	Weapon_OnSpawn(iWeapon);
	
	return iWeapon;
}

Weapon_Give(const iPlayer)
{
	if (!IsValidPev(iPlayer))
	{
		return FM_NULLENT;
	}
	
	new iWeapon, Float: vecOrigin[3];
	pev(iPlayer, pev_origin, vecOrigin);
	
	if ((iWeapon = Weapon_Create(vecOrigin)) != FM_NULLENT)
	{
		Player_DropWeapons(iPlayer, ExecuteHamB(Ham_Item_ItemSlot, iWeapon));
		
		set_pev(iWeapon, pev_spawnflags, pev(iWeapon, pev_spawnflags) | SF_NORESPAWN);
		MDLL_Touch(iWeapon, iPlayer);
		
		SetAmmoInventory(iPlayer, PrimaryAmmoIndex(iWeapon), WEAPON_DEFAULT_AMMO);
		
		return iWeapon;
	}
	
	return FM_NULLENT;
}

Player_DropWeapons(const iPlayer, const iSlot)
{
	new szWeaponName[32], iItem = get_pdata_cbase(iPlayer, m_rgpPlayerItems_CBasePlayer + iSlot, extra_offset_player);

	while (IsValidPev(iItem))
	{
		pev(iItem, pev_classname, szWeaponName, charsmax(szWeaponName));
		engclient_cmd(iPlayer, "drop", szWeaponName);

		iItem = get_pdata_cbase(iItem, m_pNext, extra_offset_weapon);
	}
}

Weapon_SendAnim(const iPlayer, const iAnim)
{
	set_pev(iPlayer, pev_weaponanim, iAnim);

	message_begin(MSG_ONE_UNRELIABLE, SVC_WEAPONANIM, {0.0, 0.0, 0.0}, iPlayer);
	write_byte(iAnim);
	write_byte(0);
	message_end();
}

stock Weapon_DefaultDeploy(const iItem, const iPlayer, const szViewModel[], const szWeaponModel[], const iAnim, const szAnimExt[])
{
	set_pev(iPlayer, pev_viewmodel2, szViewModel);
	set_pev(iPlayer, pev_weaponmodel2, szWeaponModel);
	set_pev(iPlayer, pev_fov, 90.0);
	
	set_pdata_int(iPlayer, m_iFOV, 90, extra_offset_player);
	set_pdata_int(iPlayer, m_fResumeZoom, 0, extra_offset_player);
	set_pdata_int(iPlayer, m_iLastZoom, 90, extra_offset_player);
	
	set_pdata_string(iPlayer, m_szAnimExtention * 4, szAnimExt, -1, extra_offset_player * 4);
	
	// Delay for blocking UpdateClientData
	set_pdata_float(iItem, m_flLastEventCheck, get_gametime() + 0.2, extra_offset_weapon);
	
	Weapon_SendAnim(iPlayer, iAnim);
	
}



#define BitSet(%0,%1) (%0 |= (1 << (%1 - 1)))
#define BitClear(%0,%1) (%0 &= ~(1 << (%1 - 1)))
#define BitCheck(%0,%1) (%0 & (1 << (%1 - 1)))

public client_putinserver(id)
{
	BitSet(g_bitIsConnected, id);
}

public client_disconnected(id)
{
	BitClear(g_bitIsConnected, id);
}


bool: CheckItem(const iItem, &iPlayer)
{
	if (!IsValidPev(iItem) || !IsCustomItem(iItem))
	{
		return false;
	}
	
	iPlayer = get_pdata_cbase(iItem, m_pPlayer, extra_offset_weapon);
	
	if (!IsValidPev(iPlayer) || !BitCheck(g_bitIsConnected, iPlayer))
	{
		return false;
	}
	
	return true;
}

bool: CheckItem2(const iPlayer, &iItem)
{
	if (!IsValidPev(iPlayer) || zp_get_user_zombie(iPlayer) || !BitCheck(g_bitIsConnected, iPlayer) || zp_get_user_nemesis(iPlayer))
	{
		return false;
	}
	
	iItem = get_pdata_cbase(iPlayer, m_pActiveItem, extra_offset_player);
	
	if (!IsValidPev(iItem) || !IsCustomItem(iItem))
	{
		return false;
	}
	
	return true;
}

//**********************************************
//* Some usefull stocks.                       *
//**********************************************

stock FindHullIntersection(const Float: vecSrc[3], &iTrace, const Float: vecMins[3], const Float: vecMaxs[3], const iEntity)
{
	new iTempTrace;
	
	new Float: flFraction;
	new Float: flThisDistance;
	
	new Float: vecEnd[3];
	new Float: vecEndPos[3];
	new Float: vecHullEnd[3];
	new Float: vecMinMaxs[2][3];
	
	new Float: flDistance = 999999.0;
	
	xs_vec_copy(vecMins, vecMinMaxs[0]);
	xs_vec_copy(vecMaxs, vecMinMaxs[1]);
	
	get_tr2(iTrace, TR_vecEndPos, vecHullEnd);
	
	xs_vec_sub(vecHullEnd, vecSrc, vecHullEnd);
	xs_vec_mul_scalar(vecHullEnd, 2.0, vecHullEnd);
	xs_vec_add(vecHullEnd, vecSrc, vecHullEnd);
	
	engfunc(EngFunc_TraceLine, vecSrc, vecHullEnd, DONT_IGNORE_MONSTERS, iEntity, (iTempTrace = create_tr2()));
	get_tr2(iTempTrace, TR_flFraction, flFraction);
	
	if (flFraction < 1.0)
	{
		free_tr2(iTrace);
		
		iTrace = iTempTrace;
		return;
	}
	
	for (new j, k, i = 0; i < 2; i++)
	{
		for (j = 0; j < 2; j++)
		{
			for (k = 0; k < 2; k++)
			{
				vecEnd[0] = vecHullEnd[0] + vecMinMaxs[i][0];
				vecEnd[1] = vecHullEnd[1] + vecMinMaxs[j][1];
				vecEnd[2] = vecHullEnd[2] + vecMinMaxs[k][2];
				
				engfunc(EngFunc_TraceLine, vecSrc, vecEnd, DONT_IGNORE_MONSTERS, iEntity, iTempTrace);
				get_tr2(iTempTrace, TR_flFraction, flFraction);
				
				if (flFraction < 1.0)
				{
					get_tr2(iTempTrace, TR_vecEndPos, vecEndPos);
					xs_vec_sub(vecEndPos, vecSrc, vecEndPos);
					
					if ((flThisDistance = xs_vec_len(vecEndPos)) < flDistance)
					{
						free_tr2(iTrace);
						
						iTrace = iTempTrace;
						flDistance = flThisDistance;
					}
				}
			}
		}
	}
}

stock Weapon_GetGunPosition(const iPlayer, Float: fOrigin[3], Float: fAngles[3], Float: add_forward = 0.0, Float: add_right = 0.0, Float: add_up = 0.0)
{
	static Float: Forward[3], Float: Right[3], Float: Up[3];

	ExecuteHamB(Ham_Player_GetGunPosition, iPlayer, fOrigin);
	
	pev(iPlayer, pev_angles, fAngles);
	pev(iPlayer, pev_v_angle, fAngles);
	
	global_get(glb_v_forward, Forward);
	global_get(glb_v_right, Right);
	global_get(glb_v_up, Up);
	
	xs_vec_mul_scalar(Forward, add_forward, Forward);
	xs_vec_mul_scalar(Right, add_right, Right);
	xs_vec_mul_scalar(Up, add_up, Up);
	
	fOrigin[0] = fOrigin[0] + Forward[0] + Right[0] + Up[0];
	fOrigin[1] = fOrigin[1] + Forward[1] + Right[1] + Up[1];
	fOrigin[2] = fOrigin[2] + Forward[2] + Right[2] + Up[2];
}

stock Player_SetAnimation(const iPlayer, const szAnim[])
{
   #define ACT_RANGE_ATTACK1   28
   
   // Linux extra offsets
   #define extra_offset_animating   4
   
   // CBaseAnimating
   #define m_flFrameRate      36
   #define m_flGroundSpeed      37
   #define m_flLastEventCheck   38
   #define m_fSequenceFinished   39
   #define m_fSequenceLoops   40
   
   // CBaseMonster
   #define m_Activity      73
   #define m_IdealActivity      74
   
   // CBasePlayer
   #define m_flLastAttackTime   220
   
   new iAnimDesired, Float: flFrameRate, Float: flGroundSpeed, bool: bLoops;
      
   if ((iAnimDesired = lookup_sequence(iPlayer, szAnim, flFrameRate, bLoops, flGroundSpeed)) == -1)
   {
      iAnimDesired = 0;
   }
   
   new Float: flGametime = get_gametime();

   set_pev(iPlayer, pev_frame, 0.0);
   set_pev(iPlayer, pev_framerate, 1.0);
   set_pev(iPlayer, pev_animtime, flGametime );
   set_pev(iPlayer, pev_sequence, iAnimDesired);
   
   set_pdata_int(iPlayer, m_fSequenceLoops, bLoops, extra_offset_animating);
   set_pdata_int(iPlayer, m_fSequenceFinished, 0, extra_offset_animating);
   
   set_pdata_float(iPlayer, m_flFrameRate, flFrameRate, extra_offset_animating);
   set_pdata_float(iPlayer, m_flGroundSpeed, flGroundSpeed, extra_offset_animating);
   set_pdata_float(iPlayer, m_flLastEventCheck, flGametime , extra_offset_animating);
   
   set_pdata_int(iPlayer, m_Activity, ACT_RANGE_ATTACK1, extra_offset_player);
   set_pdata_int(iPlayer, m_IdealActivity, ACT_RANGE_ATTACK1, extra_offset_player);   
   set_pdata_float(iPlayer, m_flLastAttackTime, flGametime , extra_offset_player);
}


//**********************************************
//* Weapon list update.                        *
//**********************************************

public HamHook_Item_AddToPlayer(const iItem, const iPlayer)
{
	if (!IsValidPev(iItem) || !IsValidPev(iPlayer))
	{
		return HAM_IGNORED;
	}
	
	MsgHook_WeaponList(MSGID_WEAPONLIST, iItem, iPlayer);
	return HAM_IGNORED;
}

public MsgHook_WeaponList(const iMsgID, const iMsgDest, const iMsgEntity)
{
	static arrWeaponListData[8];
	
	if (!iMsgEntity)
	{
		new szWeaponName[32];
		get_msg_arg_string(1, szWeaponName, charsmax(szWeaponName));
		
		if (!strcmp(szWeaponName, WEAPON_REFERANCE))
		{
			for (new i, a = sizeof arrWeaponListData; i < a; i++)
			{
				arrWeaponListData[i] = get_msg_arg_int(i + 2);
			}
		}
	}
	else
	{
		if (!IsCustomItem(iMsgDest) && pev(iMsgDest, pev_impulse))
		{
			return;
		}
		
		MESSAGE_BEGIN(MSG_ONE, iMsgID, {0.0, 0.0, 0.0}, iMsgEntity);
		WRITE_STRING(IsCustomItem(iMsgDest) ? WEAPON_NAME : WEAPON_REFERANCE);
		
		for (new i, a = sizeof arrWeaponListData; i < a; i++)
		{
			WRITE_BYTE(arrWeaponListData[i]);
		}
		
		MESSAGE_END();
	}
}

//**********************************************
//* Weaponbox world model.                     *
//**********************************************

public HamHook_Weaponbox_Spawn_Post(const iWeaponBox)
{
	if (IsValidPev(iWeaponBox))
	{
		state (IsValidPev(pev(iWeaponBox, pev_owner))) WeaponBox: Enabled;
	}
	
	return HAM_IGNORED;
}

public FakeMeta_SetModel(const iEntity) <WeaponBox: Enabled>
{
	state WeaponBox: Disabled;
	
	if (!IsValidPev(iEntity))
	{
		return FMRES_IGNORED;
	}
	
	#define MAX_ITEM_TYPES	6
	
	for (new i, iItem; i < MAX_ITEM_TYPES; i++)
	{
		iItem = get_pdata_cbase(iEntity, m_rgpPlayerItems_CWeaponBox + i, extra_offset_weapon);
		
		if (IsValidPev(iItem) && IsCustomItem(iItem))
		{
			SET_MODEL(iEntity, MODEL_WORLD);
			set_pev(iEntity, pev_body, 27);
			return FMRES_SUPERCEDE;
		}
	}
	
	return FMRES_IGNORED;
}

public FakeMeta_SetModel()	</* Empty statement */>	{ /*  Fallback  */ return FMRES_IGNORED; }
public FakeMeta_SetModel() 	< WeaponBox: Disabled >	{ /* Do nothing */ return FMRES_IGNORED; }

public zp_user_infected_post(iPlayer) 
{
	remove_task(iPlayer+THINK_ATTACK);
}

//**********************************************
//* Ammo Inventory.                            *
//**********************************************

PrimaryAmmoIndex(const iItem)
{
	return get_pdata_int(iItem, m_iPrimaryAmmoType, extra_offset_weapon);
}

GetAmmoInventory(const iPlayer, const iAmmoIndex)
{
	if (iAmmoIndex == -1)
	{
		return -1;
	}

	return get_pdata_int(iPlayer, m_rgAmmo_CBasePlayer + iAmmoIndex, extra_offset_player);
}

SetAmmoInventory(const iPlayer, const iAmmoIndex, const iAmount)
{
	if (iAmmoIndex == -1)
	{
		return 0;
	}

	set_pdata_int(iPlayer, m_rgAmmo_CBasePlayer + iAmmoIndex, iAmount, extra_offset_player);
	return 1;
}