#include <amxmodx>
#include <fakemeta>
#include <fakemeta_util>
#include <hamsandwich>
#include <xs>
#include <zombieplague_x>

#define PLUGIN "[ZP] Extra: CSO Weapon"
#define VERSION "1.0"
#define AUTHOR "KORD_12.7 | qubka"

#pragma ctrlchar '\'

//**********************************************
//* Weapon Settings.                           *
//**********************************************

// Main
#define WEAPON_NAME 				"weapon_spas12_zpx"
#define WEAPON_REFERANCE			"weapon_xm1014"
#define WEAPON_MAX_CLIP				8
#define WEAPON_DEFAULT_AMMO			50

#define WEAPON_MULTIPLIER_DAMAGE1 	1.52
#define WEAPON_MULTIPLIER_DAMAGE2	1.32

#define WEAPON_TIME_NEXT_IDLE 		5.0
#define WEAPON_TIME_NEXT_ATTACK 	0.5
#define WEAPON_TIME_DELAY_DEPLOY 	1.23

#define ZP_ITEM_NAME				"SPAS12EX" 
#define ZP_ITEM_COST				25

// Models
#define MODEL_WORLD				"models/zombie_plague/w_spas12ex.mdl"
#define MODEL_VIEW				"models/zombie_plague/v_spas12ex.mdl"
#define MODEL_PLAYER			"models/zombie_plague/p_spas12ex.mdl"

// Sounds
#define SOUND_FIRE				"weapons/spas12_shoot.wav"
#define SOUND_FIRE_EX			"weapons/spas12ex_shoot_2.wav"
#define SOUND_RELOAD			"weapons/spas12_reload.wav"
#define SOUND_INSERT			"weapons/spas12_insert.wav"
#define SOUND_DRAW				"weapons/spas12_draw.wav"
#define SOUND_EX1				"weapons/spas12ex_change1.wav"
#define SOUND_EX2				"weapons/spas12ex_change2.wav"

// Sprites
#define WEAPON_HUD_TXT			"sprites/weapon_spas12_zpx.txt"
#define WEAPON_HUD_SPR_1		"sprites/zpx/640hud67.spr"

#define MODEL_SHELL				"models/shotgunshell.mdl"

// Animation
#define ANIM_EXTENSION			"shotgun"

// Animation sequences
enum
{	
	ANIM_IDLE,
	ANIM_SHOOT1,
	ANIM_SHOOT2,
	ANIM_INSERT,
	ANIM_AFTER_RELOAD,
	ANIM_START_RELOAD,
	ANIM_DRAW,
	ANIM_SHOOT_EX,
	ANIM_CHANGE_NONE,
	ANIM_CHANGE_EX,
	ANIM_IDLE_EX,
	ANIM_INSERT_EX,
	ANIM_AFTER_RELOAD_EX,
	ANIM_START_RELOAD_EX,
	ANIM_DRAW_EX
};

//**********************************************
//* Some macroses.                             *
//**********************************************

#define MDLL_Spawn(%0)			dllfunc(DLLFunc_Spawn, %0)
#define MDLL_Touch(%0,%1)		dllfunc(DLLFunc_Touch, %0, %1)

#define SET_MODEL(%0,%1)		engfunc(EngFunc_SetModel, %0, %1)
#define SET_ORIGIN(%0,%1)		engfunc(EngFunc_SetOrigin, %0, %1)

#define PRECACHE_MODEL(%0)		engfunc(EngFunc_PrecacheModel, %0)
#define PRECACHE_SOUND(%0)		engfunc(EngFunc_PrecacheSound, %0)
#define PRECACHE_GENERIC(%0)		engfunc(EngFunc_PrecacheGeneric, %0)

#define MESSAGE_BEGIN(%0,%1,%2,%3)	engfunc(EngFunc_MessageBegin, %0, %1, %2, %3)
#define MESSAGE_END()			message_end()

#define WRITE_ANGLE(%0)			engfunc(EngFunc_WriteAngle, %0)
#define WRITE_BYTE(%0)			write_byte(%0)
#define WRITE_COORD(%0)			engfunc(EngFunc_WriteCoord, %0)
#define WRITE_STRING(%0)		write_string(%0)
#define WRITE_SHORT(%0)			write_short(%0)

#define INSTANCE(%0)			((%0 == -1) ? 0 : %0)

#define BitSet(%0,%1) (%0 |= (1 << (%1 - 1)))
#define BitClear(%0,%1) (%0 &= ~(1 << (%1 - 1)))
#define BitCheck(%0,%1) (%0 & (1 << (%1 - 1)))

//**********************************************
//* PvData Offsets.                            *
//**********************************************

// Linux extra offsets
#define extra_offset_weapon		4
#define extra_offset_player		5

new g_bitIsConnected;

#define m_rgpPlayerItems_CWeaponBox	34

// CBasePlayerItem
#define m_pPlayer			41
#define m_pNext				42

// CBasePlayerWeapon
#define m_flNextPrimaryAttack		46
#define m_flNextSecondaryAttack		47
#define m_flTimeWeaponIdle		48
#define m_iPrimaryAmmoType		49
#define m_iClip					51
#define m_fInModeSwitch			30
#define m_iDirection			60
#define m_iShotsFired			64
#define m_fWeaponState			74
#define m_flLastEventCheck  	38
#define m_iLastZoom 			109
#define m_fInSpecialReload		55

// CBaseMonster
#define m_flNextAttack			83

// CBasePlayer
#define m_iFOV				363
#define m_fResumeZoom       110
#define m_rgpPlayerItems_CBasePlayer	367
#define m_pActiveItem			373
#define m_rgAmmo_CBasePlayer		376
#define m_szAnimExtention		492

#define OFFSET_M3_AMMO 			381

#define IsValidPev(%0) (pev_valid(%0) == 2)

#define IsObserver(%0) pev(%0,pev_iuser1)

#define OBS_IN_EYE 4

//**********************************************
//* Let's code our weapon.                     *
//**********************************************

Weapon_OnPrecache()
{
	PRECACHE_MODEL(MODEL_VIEW);
	PRECACHE_MODEL(MODEL_PLAYER);
	PRECACHE_MODEL(MODEL_WORLD) ;
	
	PRECACHE_MODEL(MODEL_SHELL);
	PRECACHE_SOUND(SOUND_FIRE);
	PRECACHE_SOUND(SOUND_FIRE_EX);
	
	PRECACHE_SOUND(SOUND_RELOAD);
	PRECACHE_SOUND(SOUND_INSERT);
	PRECACHE_SOUND(SOUND_DRAW);
	PRECACHE_SOUND(SOUND_EX1);
	PRECACHE_SOUND(SOUND_EX2);
	
	PRECACHE_GENERIC(WEAPON_HUD_TXT);
	PRECACHE_GENERIC(WEAPON_HUD_SPR_1);
}

Weapon_OnSpawn(const iItem)
{
	// Setting world model.
	SET_MODEL(iItem, MODEL_WORLD);
}

Weapon_OnDeploy(const iItem, const iPlayer, const iClip, const iModeSwitch, const iAmmoPrimary)
{
	#pragma unused iClip, iAmmoPrimary
	
	static iszViewModel;
	if (iszViewModel || (iszViewModel = engfunc(EngFunc_AllocString, MODEL_VIEW)))
	{
		set_pev_string(iPlayer, pev_viewmodel2, iszViewModel);
	}
	
	static iszPlayerModel;
	if (iszPlayerModel || (iszPlayerModel = engfunc(EngFunc_AllocString, MODEL_PLAYER)))
	{
		set_pev_string(iPlayer, pev_weaponmodel2, iszPlayerModel);
	}
	
	set_pdata_int(iItem, m_fInSpecialReload, 0, extra_offset_weapon);
	
	set_pdata_string(iPlayer, m_szAnimExtention * 4, ANIM_EXTENSION, -1, extra_offset_player * 4);
	
	set_pdata_float(iItem, m_flTimeWeaponIdle, WEAPON_TIME_DELAY_DEPLOY, extra_offset_weapon);
	set_pdata_float(iPlayer, m_flNextAttack, WEAPON_TIME_DELAY_DEPLOY, extra_offset_player);
	
	// Deploy with default weapon settings
	Weapon_DefaultDeploy(iItem, iPlayer, MODEL_VIEW, MODEL_PLAYER, iModeSwitch ? ANIM_DRAW_EX : ANIM_DRAW, ANIM_EXTENSION);
}

Weapon_OnHolster(const iItem, const iPlayer, const iClip, const iModeSwitch, const iAmmoPrimary)
{
	#pragma unused iPlayer, iClip, iAmmoPrimary, iModeSwitch
	
	// Cancel any reload in progress.
	set_pdata_int(iItem, m_fInSpecialReload, 0, extra_offset_weapon);
}

Weapon_OnReload(const iItem, const iPlayer, const iClip, const iModeSwitch, const iAmmoPrimary)
{
	if(iAmmoPrimary <= 0)
	{
		return HAM_IGNORED;
	}
	
	if(iClip >= WEAPON_MAX_CLIP)
	{
		return HAM_IGNORED;
	}
	
	static iReloadMode; 
	iReloadMode = get_pdata_int(iItem, m_fInSpecialReload, extra_offset_weapon);

	if(iClip <= 0) 
	{
		switch(iReloadMode)
		{
			case 0:
			{
				Weapon_SendAnim(iPlayer, iModeSwitch ? ANIM_START_RELOAD_EX: ANIM_START_RELOAD);
			
				set_pdata_float(iItem, m_flNextPrimaryAttack, 0.5, extra_offset_weapon);
				set_pdata_float(iItem, m_flTimeWeaponIdle, 0.5, extra_offset_weapon);
			
				set_pdata_int(iItem, m_fInSpecialReload, 1, extra_offset_weapon);
			
				return HAM_IGNORED;
			}
			case 1:
			{		
				if (get_pdata_int(iItem, m_flTimeWeaponIdle, extra_offset_weapon) > 0.0)
				{
					return HAM_IGNORED;
				}
				
				Weapon_SendAnim(iPlayer, iModeSwitch ? ANIM_INSERT_EX : ANIM_INSERT);
					
				set_pdata_int(iItem, m_fInSpecialReload, 2, extra_offset_weapon);
				set_pdata_float(iItem, m_flTimeWeaponIdle, 0.5, extra_offset_weapon);
				
				static szAnimation[64];

				formatex(szAnimation, charsmax(szAnimation), "ref_reload_shotgun");
				Player_SetAnimation(iPlayer, szAnimation);
				
			}
			case 2:
			{
				set_pdata_int(iItem, m_iClip, iClip + 1, extra_offset_weapon);
				set_pdata_int(iPlayer, OFFSET_M3_AMMO, iAmmoPrimary-1, extra_offset_player);
				set_pdata_int(iItem, m_fInSpecialReload, 1, extra_offset_weapon);
			}
		}
		return HAM_IGNORED;
	}
	
	switch(iReloadMode)
	{
		case 0:
		{
			Weapon_SendAnim(iPlayer, iModeSwitch ? ANIM_START_RELOAD_EX : ANIM_START_RELOAD);
			
			set_pdata_float(iItem, m_flNextPrimaryAttack, 0.5, extra_offset_weapon);
			set_pdata_float(iItem, m_flTimeWeaponIdle, 0.5, extra_offset_weapon);
			
			set_pdata_int(iItem, m_fInSpecialReload, 1, extra_offset_weapon);
			
			return HAM_IGNORED;
		}
	}
	return HAM_IGNORED;
}


Weapon_OnIdle(const iItem, const iPlayer, const iClip, const iModeSwitch, const iAmmoPrimary)
{
	#pragma unused iClip, iAmmoPrimary
	
	ExecuteHamB(Ham_Weapon_ResetEmptySound, iItem);

	static iReloadMode; 
	iReloadMode = get_pdata_int(iItem, m_fInSpecialReload, extra_offset_weapon);
	
	if(iClip == WEAPON_MAX_CLIP)
	{
		if(iReloadMode == 2)
		{
			Weapon_ReloadEnd(iItem, iPlayer, iModeSwitch);
			return;
		}
	}
	
	if(iAmmoPrimary <= 0)
	{
		if(iReloadMode == 2)
		{
			Weapon_ReloadEnd(iItem, iPlayer, iModeSwitch);
			return;
		}
	}

	switch(iReloadMode)
	{
		case 1:
		{		
			if (get_pdata_int(iItem, m_flTimeWeaponIdle, extra_offset_weapon) > 0.0)
			{
				return;
			}
			
			Weapon_SendAnim(iPlayer, iModeSwitch ? ANIM_INSERT_EX: ANIM_INSERT);
				
			set_pdata_int(iItem, m_fInSpecialReload, 2, extra_offset_weapon);
			set_pdata_float(iItem, m_flTimeWeaponIdle, 0.5, extra_offset_weapon);
			
			static szAnimation[64];

			formatex(szAnimation, charsmax(szAnimation), "ref_reload_shotgun");
			Player_SetAnimation(iPlayer, szAnimation);
			
		}
		case 2:
		{
			set_pdata_int(iItem, m_iClip, iClip + 1, extra_offset_weapon);
			set_pdata_int(iPlayer, OFFSET_M3_AMMO, iAmmoPrimary-1, extra_offset_player);
			set_pdata_int(iItem, m_fInSpecialReload, 1, extra_offset_weapon);
		}
	}
	
	if(!iReloadMode)
	{
		if (get_pdata_int(iItem, m_flTimeWeaponIdle, extra_offset_weapon) > 0.0)
		{
			return;
		}
	
		set_pdata_float(iItem, m_flTimeWeaponIdle, WEAPON_TIME_NEXT_IDLE, extra_offset_weapon);
		Weapon_SendAnim(iPlayer, iModeSwitch ? ANIM_IDLE_EX : ANIM_IDLE);
	}
}

Weapon_OnPrimaryAttack(const iItem, const iPlayer, const iClip, const iModeSwitch, const iAmmoPrimary)
{
	#pragma unused iAmmoPrimary

	static iReloadMode;
	
	iReloadMode = get_pdata_int(iItem, m_fInSpecialReload, extra_offset_weapon);
	
	if(iReloadMode > 0 && iClip > 0)
	{
		Weapon_ReloadEnd(iItem, iPlayer, iModeSwitch);
		return;
	}
	
	CallOrigFireBullets3(iItem, iPlayer);
	
	if (iClip <= 0)
	{
		return;
	}
	
	static iFlags, iAnimDesired; 
	static szAnimation[64], Float: vecVelocity[3];

	iFlags = pev(iPlayer, pev_flags);

	formatex(szAnimation, charsmax(szAnimation), "ref_shoot_%s", ANIM_EXTENSION);
	
	switch(iModeSwitch)
	{
		case 0:
		{
			Weapon_SendAnim(iPlayer, random_num(ANIM_SHOOT1, ANIM_SHOOT2));
			
			set_pdata_float(iItem, m_flTimeWeaponIdle, WEAPON_TIME_NEXT_IDLE + 0.5, extra_offset_weapon);
			set_pdata_float(iItem, m_flNextPrimaryAttack, WEAPON_TIME_NEXT_ATTACK + 0.5, extra_offset_weapon);
			set_pdata_float(iItem, m_flNextSecondaryAttack, WEAPON_TIME_NEXT_ATTACK + 0.5, extra_offset_weapon);
			
			emit_sound(iPlayer, CHAN_WEAPON, SOUND_FIRE, 0.9, ATTN_NORM, 0, PITCH_NORM);
		}
		case 1:
		{
			Weapon_SendAnim(iPlayer, ANIM_SHOOT_EX);
			
			set_pdata_float(iItem, m_flTimeWeaponIdle, WEAPON_TIME_NEXT_IDLE, extra_offset_weapon);
			set_pdata_float(iItem, m_flNextPrimaryAttack, WEAPON_TIME_NEXT_ATTACK, extra_offset_weapon);
			set_pdata_float(iItem, m_flNextSecondaryAttack, WEAPON_TIME_NEXT_ATTACK, extra_offset_weapon);
			
			emit_sound(iPlayer, CHAN_WEAPON, SOUND_FIRE_EX, 0.9, ATTN_NORM, 0, PITCH_NORM);
		}
	}
	
	static iShellModelIndex; 
	  
	if (!iShellModelIndex)
	{
		iShellModelIndex = PRECACHE_MODEL(MODEL_SHELL);
	}
	
	EjectBrass(iPlayer, iShellModelIndex, 1, .flForwardScale = 8.0);
	
	if ((iAnimDesired = lookup_sequence(iPlayer, szAnimation)) == -1)
	{
		iAnimDesired = 0;
	}
	
	pev(iPlayer, pev_velocity, vecVelocity);
	set_pev(iPlayer, pev_sequence, iAnimDesired);
	
	if (xs_vec_len(vecVelocity) > 0)
	{
		Weapon_KickBack(iItem, iPlayer, 1.5, 0.45, 0.225, 0.05, 6.5, 2.5, 7);
	}
	else if (!(iFlags & FL_ONGROUND))
	{
		Weapon_KickBack(iItem, iPlayer, 2.0, 1.0, 0.5, 0.35, 9.0, 6.0, 5);
	}
	else if (iFlags & FL_DUCKING)
	{
		Weapon_KickBack(iItem, iPlayer, 0.9, 0.35, 0.15, 0.025, 5.5, 1.5, 9);
	}
	else
	{
		Weapon_KickBack(iItem, iPlayer, 1.0, 0.375, 0.175, 0.0375, 5.75, 1.75, 8);
	}
}

// Reload Ended
Weapon_ReloadEnd(const iItem, const iPlayer, const iModeSwitch)
{
	Weapon_SendAnim(iPlayer, iModeSwitch ? ANIM_AFTER_RELOAD_EX : ANIM_AFTER_RELOAD);
	
	set_pdata_float(iItem, m_flNextPrimaryAttack, 1.0, extra_offset_weapon);
	set_pdata_float(iItem, m_flTimeWeaponIdle, 1.0, extra_offset_weapon);
	
	set_pdata_int(iItem, m_fInSpecialReload, 0, extra_offset_weapon);
}

//*********************************************************************
//*           Don't modify the code below this line unless            *
//*          	 you know _exactly_ what you are doing!!!             *
//*********************************************************************

#define MSGID_WEAPONLIST 78

new g_iszWeaponKey;
new g_iForwardDecalIndex;
new g_iItemID

#define IsCustomItem(%0) (pev(%0, pev_impulse) == g_iszWeaponKey)

public plugin_precache()
{
	Weapon_OnPrecache();
	
	g_iszWeaponKey = engfunc(EngFunc_AllocString, WEAPON_NAME);
	g_iForwardDecalIndex = register_forward(FM_DecalIndex, "FakeMeta_DecalIndex_Post", true);
	
	register_message(MSGID_WEAPONLIST, "MsgHook_WeaponList");
	register_message(get_user_msgid("DeathMsg"), "MsgHook_Death");
}

public plugin_init()
{
	g_iItemID = zp_register_extra_item("Spas-12 EX", 0, 0);
	
	register_plugin(PLUGIN, VERSION, AUTHOR);
	
	register_forward(FM_SetModel,		"FakeMeta_SetModel",			false);
	register_forward(FM_TraceLine,		"FakeMeta_TraceLine_Post",		true);
	register_forward(FM_PlaybackEvent,	"FakeMeta_PlaybackEvent",		false);
	register_forward(FM_UpdateClientData,	"FakeMeta_UpdateClientData_Post",	true);
	
	// Weaponbox
	RegisterHam(Ham_Spawn, 		"weaponbox", 		"HamHook_Weaponbox_Spawn_Post", true);
	
	// Hook and change damage to entities
	RegisterHam(Ham_TraceAttack,	"func_breakable",	"HamHook_Entity_TraceAttack", false);
	RegisterHam(Ham_TraceAttack,	"info_target", 		"HamHook_Entity_TraceAttack", false);
	RegisterHam(Ham_TraceAttack,	"player", 		"HamHook_Entity_TraceAttack", false);
	
	// Item (weapon) hooks
	RegisterHam(Ham_Item_Deploy,		WEAPON_REFERANCE, "HamHook_Item_Deploy_Post",	true);
	RegisterHam(Ham_Item_AddToPlayer,	WEAPON_REFERANCE, "HamHook_Item_AddToPlayer",	false);
	RegisterHam(Ham_Item_PostFrame,		WEAPON_REFERANCE, "HamHook_Item_PostFrame",	false);
	RegisterHam(Ham_Weapon_Reload,		WEAPON_REFERANCE, "HamHook_Item_Reload",	false);
	RegisterHam(Ham_Item_Holster,		WEAPON_REFERANCE, "HamHook_Item_Holster",	false);
	
	RegisterHam(Ham_Weapon_WeaponIdle,	WEAPON_REFERANCE, "HamHook_Item_WeaponIdle",	false);
	RegisterHam(Ham_Weapon_PrimaryAttack,	WEAPON_REFERANCE, "HamHook_Item_PrimaryAttack",	false);
	
	unregister_forward(FM_DecalIndex, g_iForwardDecalIndex, true);
}

public zp_extra_item_selected(id, itemid)
{
	if (itemid == g_iItemID)
	{
		Weapon_Give(id);
	}
}

//**********************************************
//* Block client weapon.                       *
//**********************************************

public FakeMeta_UpdateClientData_Post(const iPlayer, const iSendWeapons, const CD_Handle)
{
	static iActiveItem;iActiveItem = get_pdata_cbase(iPlayer, m_pActiveItem, extra_offset_player);
	
	if (!IsValidPev(iActiveItem) || !IsCustomItem(iActiveItem))
	{
		return FMRES_IGNORED;
	}

	set_cd(CD_Handle, CD_flNextAttack, get_gametime() + 0.001);
	return FMRES_IGNORED;
}

//**********************************************
//* Item (weapon) hooks.                       *
//**********************************************

	#define _call.%0(%1,%2) \
									\
	Weapon_On%0							\
	(								\
		%1, 							\
		%2,							\
									\
		get_pdata_int(%1, m_iClip, extra_offset_weapon),	\
		get_pdata_int(%1, m_fInModeSwitch, extra_offset_weapon),   \
							\
		GetAmmoInventory(%2, PrimaryAmmoIndex(%1))		\
	) 

	
public HamHook_Item_Deploy_Post(const iItem)
{
	new iPlayer; 
	
	if (!CheckItem(iItem, iPlayer))
	{
		return HAM_IGNORED;
	}
	
	_call.Deploy(iItem, iPlayer);
	return HAM_IGNORED;
}

public HamHook_Item_WeaponIdle(const iItem)
{
	static iPlayer; 
	
	if (!CheckItem(iItem, iPlayer))
	{
		return HAM_IGNORED;
	}

	_call.Idle(iItem, iPlayer);
	return HAM_SUPERCEDE;
}

public HamHook_Item_Reload(const iItem)
{
	static iPlayer; 
	
	if (!CheckItem(iItem, iPlayer))
	{
		return HAM_IGNORED;
	}
	
	_call.Reload(iItem, iPlayer);
	return HAM_SUPERCEDE;
}

public HamHook_Item_PrimaryAttack(const iItem)
{
	static iPlayer; 
	
	if (!CheckItem(iItem, iPlayer))
	{
		return HAM_IGNORED;
	}
	
	_call.PrimaryAttack(iItem, iPlayer);
	return HAM_SUPERCEDE;
}

public HamHook_Item_Holster(const iItem)
{
	new iPlayer; 
	
	if (!CheckItem(iItem, iPlayer))
	{
		return HAM_IGNORED;
	}
	
	set_pev(iPlayer, pev_viewmodel, 0);
	set_pev(iPlayer, pev_weaponmodel, 0);
	
	_call.Holster(iItem, iPlayer);
	return HAM_SUPERCEDE;
}

public HamHook_Item_PostFrame(const iItem)
{
	static iButton, iModeSwitch, iPlayer; 
	
	if (!CheckItem(iItem, iPlayer))
	{
		return HAM_IGNORED;
	}

	// Call secondary attack
	if ((iButton = pev(iPlayer, pev_button)) & IN_ATTACK2 
		&& get_pdata_float(iItem, m_flNextPrimaryAttack, extra_offset_weapon) <= 0.0
		&& get_pdata_int(iItem, m_fInSpecialReload, extra_offset_weapon) <= 0)
	{
		iModeSwitch = get_pdata_int(iItem, m_fInModeSwitch, extra_offset_weapon);
	
		switch(iModeSwitch)
		{
			case 0:
			{
				set_pdata_int(iItem, m_fInModeSwitch, 1, extra_offset_weapon);
				
				client_print(iPlayer, print_center, "Режим: Автоматический");
				
				Weapon_SendAnim(iPlayer, ANIM_CHANGE_NONE);
	
				set_pdata_float(iItem, m_flTimeWeaponIdle, 2.2, extra_offset_weapon);
				set_pdata_float(iItem, m_flNextPrimaryAttack, 1.9, extra_offset_weapon);
			}
			case 1:
			{
				set_pdata_int(iItem, m_fInModeSwitch, 0, extra_offset_weapon);
				
				client_print(iPlayer, print_center, "Режим: Одиночный");
				
				Weapon_SendAnim(iPlayer, ANIM_CHANGE_EX);
	
				set_pdata_float(iItem, m_flTimeWeaponIdle, 2.2, extra_offset_weapon);
				set_pdata_float(iItem, m_flNextPrimaryAttack, 1.9, extra_offset_weapon);
			}
		}
		set_pev(iPlayer, pev_button, iButton & ~IN_ATTACK2);
	}
	
	return HAM_IGNORED;
}

//**********************************************
//* Fire Bullets.                              *
//**********************************************

CallOrigFireBullets3(const iItem, const iPlayer)
{
	state FireBullets: Enabled;
	static Float: vecPuncheAngle[3];

	pev(iPlayer, pev_punchangle, vecPuncheAngle);
	ExecuteHam(Ham_Weapon_PrimaryAttack, iItem);
	set_pev(iPlayer, pev_punchangle, vecPuncheAngle);
	
	state FireBullets: Disabled;
}

public MsgHook_Death() <FireBullets: Enabled>
{
	static szTruncatedWeaponName[32];
	
	if (szTruncatedWeaponName[0] == EOS)
	{
		copy(szTruncatedWeaponName, charsmax(szTruncatedWeaponName), WEAPON_NAME);
		replace(szTruncatedWeaponName, charsmax(szTruncatedWeaponName), "weapon_", "");
	}
	
	set_msg_arg_string(4, szTruncatedWeaponName);
}

public FakeMeta_PlaybackEvent() <FireBullets: Enabled>
{
	return FMRES_SUPERCEDE;
}

public FakeMeta_TraceLine_Post(const Float: vecTraceStart[3], const Float: vecTraceEnd[3], const fNoMonsters, const iEntToSkip, const iTrace) <FireBullets: Enabled>
{
	static Float: vecEndPos[3];
	
	get_tr2(iTrace, TR_vecEndPos, vecEndPos);
	engfunc(EngFunc_TraceLine, vecEndPos, vecTraceStart, fNoMonsters, iEntToSkip, 0);
	
	UTIL_GunshotDecalTrace(0);
	UTIL_GunshotDecalTrace(iTrace, true);
}

public HamHook_Entity_TraceAttack(const iEntity, const iAttacker, const Float: flDamage) <FireBullets: Enabled>
{
	static iActiveItem, iModeSwitch;

	if (!BitCheck(g_bitIsConnected, iAttacker) || !IsValidPev(iAttacker))
	{
		return;
	}
	
	iActiveItem = get_pdata_cbase(iAttacker, m_pActiveItem, extra_offset_player);
	
	if (!IsValidPev(iActiveItem))
	{
		return;
	}
	
	iModeSwitch = get_pdata_int(iActiveItem, m_fInModeSwitch, extra_offset_weapon);
	
	SetHamParamFloat(3, flDamage * (iModeSwitch ? WEAPON_MULTIPLIER_DAMAGE2 : WEAPON_MULTIPLIER_DAMAGE1));
}

public MsgHook_Death()			</* Empty statement */>		{ /* Fallback */ }
public MsgHook_Death()			<FireBullets: Disabled>		{ /* Do notning */ }

public FakeMeta_PlaybackEvent() 	</* Empty statement */>		{ return FMRES_IGNORED; }
public FakeMeta_PlaybackEvent() 	<FireBullets: Disabled>		{ return FMRES_IGNORED; }

public FakeMeta_TraceLine_Post() 	</* Empty statement */>		{ /* Fallback */ }
public FakeMeta_TraceLine_Post() 	<FireBullets: Disabled>		{ /* Do notning */ }

public HamHook_Entity_TraceAttack() 	</* Empty statement */>		{ /* Fallback */ }
public HamHook_Entity_TraceAttack() 	<FireBullets: Disabled>		{ /* Do notning */ }

//**********************************************
//* Kick back.                                 *
//**********************************************

Weapon_KickBack(const iItem, const iPlayer, Float: upBase, Float: lateralBase, const Float: upMod, const Float: lateralMod, Float: upMax, Float: lateralMax, const directionChange)
{
	static iDirection; 
	static iShotsFired; 
	
	static Float: vecPunchangle[3];
	pev(iPlayer, pev_punchangle, vecPunchangle);
	
	if ((iShotsFired = get_pdata_int(iItem, m_iShotsFired, extra_offset_weapon)) != 1)
	{
		upBase += iShotsFired * upMod;
		lateralBase += iShotsFired * lateralMod;
	}
	
	upMax *= -1.0;
	vecPunchangle[0] -= upBase;
 
	if (upMax >= vecPunchangle[0])
	{
		vecPunchangle[0] = upMax;
	}
	
	if ((iDirection = get_pdata_int(iItem, m_iDirection, extra_offset_weapon)))
	{
		vecPunchangle[1] += lateralBase;
		
		if (lateralMax < vecPunchangle[1])
		{
			vecPunchangle[1] = lateralMax;
		}
	}
	else
	{
		lateralMax *= -1.0;
		vecPunchangle[1] -= lateralBase;
		
		if (lateralMax > vecPunchangle[1])
		{
			vecPunchangle[1] = lateralMax;
		}
	}
	
	if (!random_num(0, directionChange))
	{
		set_pdata_int(iItem, m_iDirection, !iDirection, extra_offset_weapon);
	}
	
	set_pev(iPlayer, pev_punchangle, vecPunchangle);
}


//**********************************************
//* Brass ejection.                            *
//**********************************************

EjectBrass(const iPlayer, const iModelIndex, const iBounce, const Float:flUpScale = -9.0, const Float: flForwardScale = 16.0, const Float: flRightScale = 0.0)
{
	static i, msgBrass;
	
	static Float: vecUp[3]; 
	static Float: vecRight[3]; 
	static Float: vecForward[3]; 
	
	static Float: vecAngle[3];
	static Float: vecOrigin[3];
	static Float: vecViewOfs[3];
	static Float: vecVelocity[3];
	
	pev(iPlayer, pev_v_angle, vecAngle);
	pev(iPlayer, pev_punchangle, vecOrigin);
	
	xs_vec_add(vecAngle, vecOrigin, vecOrigin);
	engfunc(EngFunc_MakeVectors, vecOrigin);
	
	pev(iPlayer, pev_origin, vecOrigin);
	pev(iPlayer, pev_view_ofs, vecViewOfs);
	pev(iPlayer, pev_velocity, vecVelocity);
	
	global_get(glb_v_up, vecUp);
	global_get(glb_v_right, vecRight);
	global_get(glb_v_forward, vecForward);
	
	for (i = 0; i < 3; i++)
	{
		vecOrigin[i] = vecOrigin[i] + vecViewOfs[i] + vecForward[i] * flForwardScale + vecUp[i] * flUpScale + vecRight[i] * flRightScale;
		vecVelocity[i] = vecVelocity[i] + vecForward[i] * 25.0 + vecUp[i] * random_float(100.0, 150.0) + vecRight[i] * random_float(50.0, 70.0);
	}
	
	if (msgBrass || (msgBrass = get_user_msgid("Brass")))
	{
		MESSAGE_BEGIN(MSG_PVS, msgBrass, vecOrigin, 0);
		WRITE_BYTE(0 /* dummy */);
		WRITE_COORD(vecOrigin[0]);
		WRITE_COORD(vecOrigin[1]);
		WRITE_COORD(vecOrigin[2]);
		WRITE_COORD(0.0 /* dummy */);
		WRITE_COORD(0.0 /* dummy */);
		WRITE_COORD(0.0 /* dummy */);
		WRITE_COORD(vecVelocity[0]);
		WRITE_COORD(vecVelocity[1]);
		WRITE_COORD(vecVelocity[2]);
		WRITE_ANGLE(vecAngle[1]);
		WRITE_SHORT(iModelIndex);
		WRITE_BYTE(iBounce);
		WRITE_BYTE(0 /* dummy */);
		WRITE_BYTE(iPlayer);
		MESSAGE_END();
	}
}

//**********************************************
//* Create and check our custom weapon.        *
//**********************************************

Weapon_Create(const Float: vecOrigin[3] = {0.0, 0.0, 0.0}, const Float: vecAngles[3] = {0.0, 0.0, 0.0})
{
	new iWeapon;

	static iszAllocStringCached;
	if (iszAllocStringCached || (iszAllocStringCached = engfunc(EngFunc_AllocString, WEAPON_REFERANCE)))
	{
		iWeapon = engfunc(EngFunc_CreateNamedEntity, iszAllocStringCached);
	}
	
	if (!IsValidPev(iWeapon))
	{
		return FM_NULLENT;
	}
	
	MDLL_Spawn(iWeapon);
	SET_ORIGIN(iWeapon, vecOrigin);
	
	set_pdata_int(iWeapon, m_iClip, WEAPON_MAX_CLIP, extra_offset_weapon);
	set_pdata_int(iWeapon, m_fInModeSwitch, 0, extra_offset_weapon);
	
	set_pev_string(iWeapon, pev_classname, g_iszWeaponKey );
	set_pev(iWeapon, pev_impulse, g_iszWeaponKey);
	set_pev(iWeapon, pev_angles, vecAngles);
	
	Weapon_OnSpawn(iWeapon);
	
	return iWeapon;
}

Weapon_Give(const iPlayer)
{
	if (!IsValidPev(iPlayer))
	{
		return FM_NULLENT;
	}
	
	new iWeapon, Float: vecOrigin[3];
	pev(iPlayer, pev_origin, vecOrigin);
	
	if ((iWeapon = Weapon_Create(vecOrigin)) != FM_NULLENT)
	{
		Player_DropWeapons(iPlayer, ExecuteHamB(Ham_Item_ItemSlot, iWeapon));
		
		set_pev(iWeapon, pev_spawnflags, pev(iWeapon, pev_spawnflags) | SF_NORESPAWN);
		MDLL_Touch(iWeapon, iPlayer);

		SetAmmoInventory(iPlayer, PrimaryAmmoIndex(iWeapon), WEAPON_DEFAULT_AMMO);
		
		//client_print(iPlayer, print_center, "Для смены режима жмите правую кнопку!");

		return iWeapon;
	}
	
	return FM_NULLENT;
}

Player_DropWeapons(const iPlayer, const iSlot)
{
	new szWeaponName[32], iItem = get_pdata_cbase(iPlayer, m_rgpPlayerItems_CBasePlayer + iSlot, extra_offset_player);

	while (IsValidPev(iItem))
	{
		pev(iItem, pev_classname, szWeaponName, charsmax(szWeaponName));
		engclient_cmd(iPlayer, "drop", szWeaponName);

		iItem = get_pdata_cbase(iItem, m_pNext, extra_offset_weapon);
	}
}

Weapon_SendAnim(const iPlayer, const iAnim)
{
	set_pev(iPlayer, pev_weaponanim, iAnim);

	message_begin(MSG_ONE_UNRELIABLE, SVC_WEAPONANIM, {0.0, 0.0, 0.0}, iPlayer);
	write_byte(iAnim);
	write_byte(0);
	message_end();
}

stock Weapon_DefaultDeploy(const iItem, const iPlayer, const szViewModel[], const szWeaponModel[], const iAnim, const szAnimExt[])
{
	set_pev(iPlayer, pev_viewmodel2, szViewModel);
	set_pev(iPlayer, pev_weaponmodel2, szWeaponModel);
	set_pev(iPlayer, pev_fov, 90.0);
	
	set_pdata_int(iPlayer, m_iFOV, 90, extra_offset_player);
	set_pdata_int(iPlayer, m_fResumeZoom, 0, extra_offset_player);
	set_pdata_int(iPlayer, m_iLastZoom, 90, extra_offset_player);
	
	set_pdata_string(iPlayer, m_szAnimExtention * 4, szAnimExt, -1, extra_offset_player * 4);
	
	//this->m_flFamasBurstSpread = *(float *)gpGlobals;
	
	// Delay for blocking UpdateClientData
	set_pdata_float(iItem, m_flLastEventCheck, get_gametime() + 0.2, extra_offset_weapon);
	
	Weapon_SendAnim(iPlayer, iAnim);
	
}

#define BitSet(%0,%1) (%0 |= (1 << (%1 - 1)))
#define BitClear(%0,%1) (%0 &= ~(1 << (%1 - 1)))
#define BitCheck(%0,%1) (%0 & (1 << (%1 - 1)))

public client_putinserver(id)
{
	BitSet(g_bitIsConnected, id);
}

public client_disconnected(id)
{
	BitClear(g_bitIsConnected, id);
}


bool: CheckItem(const iItem, &iPlayer)
{
	if (!IsValidPev(iItem) || !IsCustomItem(iItem))
	{
		return false;
	}
	
	iPlayer = get_pdata_cbase(iItem, m_pPlayer, extra_offset_weapon);
	
	if (!IsValidPev(iPlayer) || !BitCheck(g_bitIsConnected, iPlayer))
	{
		return false;
	}
	
	return true;
}


//**********************************************
//* Decals.                                    *
//**********************************************

new Array: g_hDecals;

public FakeMeta_DecalIndex_Post()
{
	if (!g_hDecals)
	{
		g_hDecals = ArrayCreate(1, 1);
	}
	
	ArrayPushCell(g_hDecals, get_orig_retval());
}

UTIL_GunshotDecalTrace(const iTrace, const bool: bIsGunshot = false)
{
	static iHit;
	static iMessage;
	static iDecalIndex;
	
	static Float: flFraction; 
	static Float: vecEndPos[3];
	
	iHit = INSTANCE(get_tr2(iTrace, TR_pHit));
	
	if (iHit  && !IsValidPev(iHit) || (pev(iHit, pev_flags) & FL_KILLME))
	{
		return;
	}
	
	if (pev(iHit, pev_solid) != SOLID_BSP && pev(iHit, pev_movetype) != MOVETYPE_PUSHSTEP)
	{
		return;
	}
	
	iDecalIndex = ExecuteHamB(Ham_DamageDecal, iHit, 0);
	
	if (iDecalIndex < 0 || iDecalIndex >=  ArraySize(g_hDecals))
	{
		return;
	}
	
	iDecalIndex = ArrayGetCell(g_hDecals, iDecalIndex);
	
	get_tr2(iTrace, TR_flFraction, flFraction);
	get_tr2(iTrace, TR_vecEndPos, vecEndPos);
	
	if (iDecalIndex < 0 || flFraction >= 1.0)
	{
		return;
	}
	
	if (bIsGunshot)
	{
		iMessage = TE_GUNSHOTDECAL;
	}
	else
	{
		iMessage = TE_DECAL;
		
		if (iHit != 0)
		{
			if (iDecalIndex > 255)
			{
				iMessage = TE_DECALHIGH;
				iDecalIndex -= 256;
			}
		}
		else
		{
			iMessage = TE_WORLDDECAL;
			
			if (iDecalIndex > 255)
			{
				iMessage = TE_WORLDDECALHIGH;
				iDecalIndex -= 256;
			}
		}
	}
	
	MESSAGE_BEGIN(MSG_PAS, SVC_TEMPENTITY, vecEndPos, 0);
	WRITE_BYTE(iMessage);
	WRITE_COORD(vecEndPos[0]);
	WRITE_COORD(vecEndPos[1]);
	WRITE_COORD(vecEndPos[2]);
	
	if (bIsGunshot)
	{
		WRITE_SHORT(iHit);
		WRITE_BYTE(iDecalIndex);
	}
	else 
	{
		WRITE_BYTE(iDecalIndex);
		
		if (iHit)
		{
			WRITE_SHORT(iHit);
		}
	}
    
	MESSAGE_END();
}

//**********************************************
//* Weapon list update.                        *
//**********************************************

public HamHook_Item_AddToPlayer(const iItem, const iPlayer)
{
	if (!IsValidPev(iItem) || !IsValidPev(iPlayer))
	{
		return HAM_IGNORED;
	}
	
	MsgHook_WeaponList(MSGID_WEAPONLIST, iItem, iPlayer);
	return HAM_IGNORED;
}

public MsgHook_WeaponList(const iMsgID, const iMsgDest, const iMsgEntity)
{
	static arrWeaponListData[8];
	
	if (!iMsgEntity)
	{
		new szWeaponName[32];
		get_msg_arg_string(1, szWeaponName, charsmax(szWeaponName));
		
		if (!strcmp(szWeaponName, WEAPON_REFERANCE))
		{
			for (new i, a = sizeof arrWeaponListData; i < a; i++)
			{
				arrWeaponListData[i] = get_msg_arg_int(i + 2);
			}
		}
	}
	else
	{
		if (!IsCustomItem(iMsgDest) && pev(iMsgDest, pev_impulse))
		{
			return;
		}
		
		MESSAGE_BEGIN(MSG_ONE, iMsgID, {0.0, 0.0, 0.0}, iMsgEntity);
		WRITE_STRING(IsCustomItem(iMsgDest) ? WEAPON_NAME : WEAPON_REFERANCE);
		
		for (new i, a = sizeof arrWeaponListData; i < a; i++)
		{
			WRITE_BYTE(arrWeaponListData[i]);
		}
		
		MESSAGE_END();
	}
}

//**********************************************
//* Weaponbox world model.                     *
//**********************************************

public HamHook_Weaponbox_Spawn_Post(const iWeaponBox)
{
	if (IsValidPev(iWeaponBox))
	{
		state (IsValidPev(pev(iWeaponBox, pev_owner))) WeaponBox: Enabled;
	}
	
	return HAM_IGNORED;
}

public FakeMeta_SetModel(const iEntity) <WeaponBox: Enabled>
{
	state WeaponBox: Disabled;
	
	if (!IsValidPev(iEntity))
	{
		return FMRES_IGNORED;
	}
	
	#define MAX_ITEM_TYPES	6
	
	for (new i, iItem; i < MAX_ITEM_TYPES; i++)
	{
		iItem = get_pdata_cbase(iEntity, m_rgpPlayerItems_CWeaponBox + i, extra_offset_weapon);
		
		if (IsValidPev(iItem) && IsCustomItem(iItem))
		{
			SET_MODEL(iEntity, MODEL_WORLD);
			set_pev(iEntity, pev_body, 10);
			return FMRES_SUPERCEDE;
		}
	}
	
	return FMRES_IGNORED;
}

public FakeMeta_SetModel()	</* Empty statement */>	{ /*  Fallback  */ return FMRES_IGNORED; }
public FakeMeta_SetModel() 	< WeaponBox: Disabled >	{ /* Do nothing */ return FMRES_IGNORED; }

//**********************************************
//* Ammo Inventory.                            *
//**********************************************

PrimaryAmmoIndex(const iItem)
{
	return get_pdata_int(iItem, m_iPrimaryAmmoType, extra_offset_weapon);
}

GetAmmoInventory(const iPlayer, const iAmmoIndex)
{
	if (iAmmoIndex == -1)
	{
		return -1;
	}

	return get_pdata_int(iPlayer, m_rgAmmo_CBasePlayer + iAmmoIndex, extra_offset_player);
}

SetAmmoInventory(const iPlayer, const iAmmoIndex, const iAmount)
{
	if (iAmmoIndex == -1)
	{
		return 0;
	}

	set_pdata_int(iPlayer, m_rgAmmo_CBasePlayer + iAmmoIndex, iAmount, extra_offset_player);
	return 1;
}

stock Player_SetAnimation(const iPlayer, const szAnim[])
{
	if(!is_user_alive(iPlayer))
		return;
		
   #define ACT_RANGE_ATTACK1   28
   
   // Linux extra offsets
   #define extra_offset_animating   4
   
   // CBaseAnimating
   #define m_flFrameRate      36
   #define m_flGroundSpeed      37
   #define m_flLastEventCheck   38
   #define m_fSequenceFinished   39
   #define m_fSequenceLoops   40
   
   // CBaseMonster
   #define m_Activity      73
   #define m_IdealActivity      74
   
   // CBasePlayer
   #define m_flLastAttackTime   220
   
   new iAnimDesired, Float: flFrameRate, Float: flGroundSpeed, bool: bLoops;
      
   if ((iAnimDesired = lookup_sequence(iPlayer, szAnim, flFrameRate, bLoops, flGroundSpeed)) == -1)
   {
      iAnimDesired = 0;
   }
   
   new Float: flGametime = get_gametime();

   set_pev(iPlayer, pev_frame, 0.0);
   set_pev(iPlayer, pev_framerate, 1.0);
   set_pev(iPlayer, pev_animtime, flGametime );
   set_pev(iPlayer, pev_sequence, iAnimDesired);
   
   set_pdata_int(iPlayer, m_fSequenceLoops, bLoops, extra_offset_animating);
   set_pdata_int(iPlayer, m_fSequenceFinished, 0, extra_offset_animating);
   
   set_pdata_float(iPlayer, m_flFrameRate, flFrameRate, extra_offset_animating);
   set_pdata_float(iPlayer, m_flGroundSpeed, flGroundSpeed, extra_offset_animating);
   set_pdata_float(iPlayer, m_flLastEventCheck, flGametime , extra_offset_animating);
   
   set_pdata_int(iPlayer, m_Activity, ACT_RANGE_ATTACK1, extra_offset_player);
   set_pdata_int(iPlayer, m_IdealActivity, ACT_RANGE_ATTACK1, extra_offset_player);   
   set_pdata_float(iPlayer, m_flLastAttackTime, flGametime , extra_offset_player);
}
